#include "sys.h"
#include "lex.h"

/* #define DEBUG FALSE */

typedef char *cell_ptr;                 /* this module has no knowledge of */
                                        /*  spreadsheet structure          */
#define BUFF_LENGTH 255

token_type token;
union token_value tok_val;
static char line_buffer[BUFF_LENGTH];
static char *buf_ptr;


void init_lex()
{
        buf_ptr = line_buffer;
}


#ifdef DEBUG
token_type yylex()
{
        printf("zyylex, ");
        zyylex();
        printf("*tok_val = %lf, token = %d\n", *(tok_val.ptr), token);
        return token;
}

token_type zyylex()
#else
token_type yylex()
#endif
{
        double getdoub();

        extern cell_ptr spread[NUM_ROW][NUM_COL];

        cell_ptr cellalloc();
        DATA_TYPE *ret_val_ptr();

        char *ptr, tok_sym[MAX_SYM_LEN];
        int col, row;
        int ch, aux;
        cell_ptr cell_loc;

        while(isspace( ch = next_buf_char() ));

        if(isalpha(ch))
        {
                ptr = tok_sym;          /* get symbol into buffer tok_sym */
                *ptr = toupper(ch);
                ptr++;
                while( isalpha(ch = next_buf_char() ) || isdigit(ch))
                         *(ptr++) = toupper(ch);
                if(ch != EOS)unnext_buf_char();
                *ptr = EOS;

                aux = is_keyword(tok_sym);
                if(aux)
                {
                        token = aux;
                        return token;
                }
                col = extr_col(tok_sym);
                if(col < NUM_COL && col >= 0)
                {
                        row = extr_row(tok_sym);
                        if(row < NUM_ROW && row >= 0)
                        {
                                cell_loc = spread[row][col];
                                if(! cell_loc)
                                    cell_loc = spread[row][col] = cellalloc();
                                token = IDENT_TOK;
                                tok_val.ptr = ret_val_ptr(cell_loc);
                                return IDENT_TOK;
                        }
                }
        }

        if(isdigit(ch) || ch == '.')
        {
                unnext_buf_char();
                tok_val.val = getdoub();

                token = NUMBER;
                return NUMBER;
        }

        if(ch == EOS)
        {
                token = EOLN;
                return EOLN;
        }


        token = ch;
        return ch;
}



char *last_line()
{
        return line_buffer;
}

int extr_row(s)
char *s;
{
        int accum;

        accum = 0;

        while(isalpha(*s)) s++;

        while(isdigit(*s))
        {
                accum = accum * 10 + (*s - '0');
                s++;
        }
        if(*s != EOS) accum = -1 + FIRST_ROW;
        return accum - FIRST_ROW;
}


int extr_col(s)
char *s;
{
        int accum;

        accum = -1 + FIRST_COL;

        if(isalpha(*s)) accum = *s;
        s++;
        if(! isdigit(*s) && *s != EOS) accum = -1 + FIRST_COL;

        return accum - FIRST_COL;
}


char next_buf_char()
{
        char ch;

        ch = *buf_ptr;
        if(ch != EOS) buf_ptr++;

        return ch;
}

void unnext_buf_char()
{
        if(buf_ptr != line_buffer)buf_ptr--;
}


int complete()
{
        return(*buf_ptr == EOS);
}


int user_input()
{
        int ch;

        while((ch = inkey()) == 0);
        if(isalnum(ch) || ch == '-' || ch == '(' || ch == '"'
        || ch == '+'   || ch == '.')
        {
                if(getline(ch))
                {
                        init_lex();
                        if(ch == '"') return LABEL;

                        return EXPR;
                }
        }
        return ch;
}

#define BEEP 7
#define BSP  8
#define CR  13

#define PRMPT_STR "> "
#define PRMPT_ROW 23
#define PRMPT_COL 2

#define BUF_LEN 75


int getline(ch)
int ch;
{
        void scr_pos(), putchar();

        char *lim;
        char getline_buffer[BUF_LEN], *pres_glbuf;
        int temp;
        unsigned int i;

        lim = &(getline_buffer[BUF_LEN - 1]);
        pres_glbuf = getline_buffer;
        *pres_glbuf = ch;
        pres_glbuf++;

        scr_pos(PRMPT_ROW, PRMPT_COL);
        printf("%s%c", PRMPT_STR, ch);

        do
        {
                while((temp = inkey()) == 0);
                if(temp == BSP)
                {
                        if(pres_glbuf == getline_buffer)
                        {
                                putchar(BEEP);
                                for(i=0; i<strlen(PRMPT_STR); i++)
                                {       /* erase prompt */
                                        putchar(BSP);
                                        putchar(' ');
                                        putchar(BSP);
                                }
                                return FALSE;   /* no line received */
                        }
                        else
                        {
                                putchar(BSP);
                                putchar(' ');
                                putchar(BSP);
                                pres_glbuf--;
                        }
                }
                else if(isprint(temp))
                {
                        if(pres_glbuf == lim)
                        {
                                putchar(BEEP);
                        }
                        else
                        {
                                putchar(temp);
                                *pres_glbuf = temp;
                                pres_glbuf++;
                        }
                }
        }while(temp != CR);

        *pres_glbuf = EOS;
        strcpy(line_buffer, getline_buffer);

        return TRUE; /* line received from user */
}


#define scr_wid 80
#define reg_attrib 7

void erase_user_input()
{
        scr_pos(PRMPT_ROW, PRMPT_COL);
        scr_wac(' ', reg_attrib, scr_wid);
}


#define TEXT_ROW PRMPT_ROW - 1
#define MAX_TXT_LEN 75

void show_text(txt, r, c)
unsigned int r;
char c;
char *txt;
{
        void scr_pos(), scr_wc();

        scr_pos(TEXT_ROW, PRMPT_COL);
        scr_wc(' ', MAX_TXT_LEN);

        printf("%c%d FORM: %s", c, r, txt);
}


typedef int jmp_buf[3];

void error(s)
char *s;
{
        char *ptr;
        unsigned int i;

        extern jmp_buf env;

        scr_pos(PRMPT_ROW+1, PRMPT_COL);

        for(i = 0; i < strlen(PRMPT_STR); i++) putchar(' ');
        for(ptr = line_buffer; ptr != buf_ptr; ptr++) putchar('-');
        printf("^");
        printf("   %s, Error occurs at or before marker", s);
        while(inkey() == 0);

        scr_pos(PRMPT_ROW+1, PRMPT_COL);
        scr_wc(' ', MAX_TXT_LEN);
        longjmp(env, 0);
}

/*
        Is char string s a keyword?  If not, return 0, otherwise, return
        the token code.

        Presently works via a direct comparison scheme.  If more keywords
        needed, a table lookup would be better.
*/

int is_keyword(s)
{
        if(strcmp(s, "QUIT") == 0) return QUIT_TOK;

        return 0;
}
